/*!
 * pixi-cjk - v0.2.4
 * Compiled Tue, 02 Mar 2021 11:38:09 UTC
 *
 * pixi-cjk is licensed under the MIT License.
 * http://www.opensource.org/licenses/mit-license
 */

(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? factory(require('pixi.js')) :
  typeof define === 'function' && define.amd ? define(['pixi.js'], factory) :
  (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global.PIXI));
}(this, (function (PIXI) { 'use strict';

  if (typeof PIXI === 'undefined') { throw 'PixiJS is required'; }

  const HIGH_SURROGATE_START = 0xd800;
  const HIGH_SURROGATE_END = 0xdbff;

  const LOW_SURROGATE_START = 0xdc00;

  const REGIONAL_INDICATOR_START = 0x1f1e6;
  const REGIONAL_INDICATOR_END = 0x1f1ff;

  const FITZPATRICK_MODIFIER_START = 0x1f3fb;
  const FITZPATRICK_MODIFIER_END = 0x1f3ff;

  const VARIATION_MODIFIER_START = 0xfe00;
  const VARIATION_MODIFIER_END = 0xfe0f;

  const DIACRITICAL_MARKS_START = 0x20d0;
  const DIACRITICAL_MARKS_END = 0x20ff;

  const ZWJ = 0x200d;

  const GRAPHEMS = [
    0x0308, // ( ◌̈ ) COMBINING DIAERESIS
    0x0937, // ( ष ) DEVANAGARI LETTER SSA
    0x0937, // ( ष ) DEVANAGARI LETTER SSA
    0x093F, // ( ि ) DEVANAGARI VOWEL SIGN I
    0x093F, // ( ि ) DEVANAGARI VOWEL SIGN I
    0x0BA8, // ( ந ) TAMIL LETTER NA
    0x0BBF, // ( ி ) TAMIL VOWEL SIGN I
    0x0BCD, // ( ◌்) TAMIL SIGN VIRAMA
    0x0E31, // ( ◌ั ) THAI CHARACTER MAI HAN-AKAT
    0x0E33, // ( ำ ) THAI CHARACTER SARA AM
    0x0E40, // ( เ ) THAI CHARACTER SARA E
    0x0E49, // ( เ ) THAI CHARACTER MAI THO
    0x1100, // ( ᄀ ) HANGUL CHOSEONG KIYEOK
    0x1161, // ( ᅡ ) HANGUL JUNGSEONG A
    0x11A8 // ( ᆨ ) HANGUL JONGSEONG KIYEOK
  ];

  function runes (string) {
    if (typeof string !== 'string') {
      throw new Error('string cannot be undefined or null')
    }
    const result = [];
    let i = 0;
    let increment = 0;
    while (i < string.length) {
      increment += nextUnits(i + increment, string);
      if (isGraphem(string[i + increment])) {
        increment++;
      }
      if (isVariationSelector(string[i + increment])) {
        increment++;
      }
      if (isDiacriticalMark(string[i + increment])) {
        increment++;
      }
      if (isZeroWidthJoiner(string[i + increment])) {
        increment++;
        continue
      }
      result.push(string.substring(i, i + increment));
      i += increment;
      increment = 0;
    }
    return result
  }

  // Decide how many code units make up the current character.
  // BMP characters: 1 code unit
  // Non-BMP characters (represented by surrogate pairs): 2 code units
  // Emoji with skin-tone modifiers: 4 code units (2 code points)
  // Country flags: 4 code units (2 code points)
  // Variations: 2 code units
  function nextUnits (i, string) {
    const current = string[i];
    // If we don't have a value that is part of a surrogate pair, or we're at
    // the end, only take the value at i
    if (!isFirstOfSurrogatePair(current) || i === string.length - 1) {
      return 1
    }

    const currentPair = current + string[i + 1];
    let nextPair = string.substring(i + 2, i + 5);

    // Country flags are comprised of two regional indicator symbols,
    // each represented by a surrogate pair.
    // See http://emojipedia.org/flags/
    // If both pairs are regional indicator symbols, take 4
    if (isRegionalIndicator(currentPair) && isRegionalIndicator(nextPair)) {
      return 4
    }

    // If the next pair make a Fitzpatrick skin tone
    // modifier, take 4
    // See http://emojipedia.org/modifiers/
    // Technically, only some code points are meant to be
    // combined with the skin tone modifiers. This function
    // does not check the current pair to see if it is
    // one of them.
    if (isFitzpatrickModifier(nextPair)) {
      return 4
    }
    return 2
  }

  function isFirstOfSurrogatePair (string) {
    return string && betweenInclusive(string[0].charCodeAt(0), HIGH_SURROGATE_START, HIGH_SURROGATE_END)
  }

  function isRegionalIndicator (string) {
    return betweenInclusive(codePointFromSurrogatePair(string), REGIONAL_INDICATOR_START, REGIONAL_INDICATOR_END)
  }

  function isFitzpatrickModifier (string) {
    return betweenInclusive(codePointFromSurrogatePair(string), FITZPATRICK_MODIFIER_START, FITZPATRICK_MODIFIER_END)
  }

  function isVariationSelector (string) {
    return typeof string === 'string' && betweenInclusive(string.charCodeAt(0), VARIATION_MODIFIER_START, VARIATION_MODIFIER_END)
  }

  function isDiacriticalMark (string) {
    return typeof string === 'string' && betweenInclusive(string.charCodeAt(0), DIACRITICAL_MARKS_START, DIACRITICAL_MARKS_END)
  }

  function isGraphem (string) {
    return typeof string === 'string' && GRAPHEMS.indexOf(string.charCodeAt(0)) !== -1
  }

  function isZeroWidthJoiner (string) {
    return typeof string === 'string' && string.charCodeAt(0) === ZWJ
  }

  function codePointFromSurrogatePair (pair) {
    const highOffset = pair.charCodeAt(0) - HIGH_SURROGATE_START;
    const lowOffset = pair.charCodeAt(1) - LOW_SURROGATE_START;
    return (highOffset << 10) + lowOffset + 0x10000
  }

  function betweenInclusive (value, lower, upper) {
    return value >= lower && value <= upper
  }

  function substring (string, start, width) {
    const chars = runes(string);
    if (start === undefined) {
      return string
    }
    if (start >= chars.length) {
      return ''
    }
    const rest = chars.length - start;
    const stringWidth = width === undefined ? rest : width;
    let endIndex = start + stringWidth;
    if (endIndex > (start + rest)) {
      endIndex = undefined;
    }
    return chars.slice(start, endIndex).join('')
  }

  var runes_1 = runes;
  var substr = substring;
  runes_1.substr = substr;

  // Line breaking rules in CJK (Kinsoku Shori)
  // Refer from https://en.wikipedia.org/wiki/Line_breaking_rules_in_East_Asian_languages
  const regexCannotStartZhCn = /[!%),.:;?\]}¢°·'""†‡›℃∶、。〃〆〕〗〞﹚﹜！＂％＇），．：；？！］｝～]/;
  const regexCannotEndZhCn = /[$(£¥·'"〈《「『【〔〖〝﹙﹛＄（．［｛￡￥]/;
  const regexCannotStartZhTw = /[!),.:;?\]}¢·–—'"•"、。〆〞〕〉》」︰︱︲︳﹐﹑﹒﹓﹔﹕﹖﹘﹚﹜！），．：；？︶︸︺︼︾﹀﹂﹗］｜｝､]/;
  const regexCannotEndZhTw = /[([{£¥'"‵〈《「『〔〝︴﹙﹛（｛︵︷︹︻︽︿﹁﹃﹏]/;
  const regexCannotStartJaJp = /[)\]｝〕〉》」』】〙〗〟'"｠»ヽヾーァィゥェォッャュョヮヵヶぁぃぅぇぉっゃゅょゎゕゖㇰㇱㇲㇳㇴㇵㇶㇷㇸㇹㇺㇻㇼㇽㇾㇿ々〻‐゠–〜?!‼⁇⁈⁉・、:;,。.]/;
  const regexCannotEndJaJp = /[([｛〔〈《「『【〘〖〝'"｟«—...‥〳〴〵]/;
  const regexCannotStartKoKr = /[!%),.:;?\]}¢°'"†‡℃〆〈《「『〕！％），．：；？］｝]/;
  const regexCannotEndKoKr = /[$([{£¥'"々〇〉》」〔＄（［｛｠￥￦#]/;
  const regexCannotStart = new RegExp(`${regexCannotStartZhCn.source}|${regexCannotStartZhTw.source}|${regexCannotStartJaJp.source}|${regexCannotStartKoKr.source}`);
  const regexCannotEnd = new RegExp(`${regexCannotEndZhCn.source}|${regexCannotEndZhTw.source}|${regexCannotEndJaJp.source}|${regexCannotEndKoKr.source}`);
  const shouldBreakByKinsokuShorui = (char, nextChar) => {
      // @ts-ignore
      const isBreakingSpace = PIXI.TextMetrics.isBreakingSpace;
      if (isBreakingSpace(nextChar))
          return false;
      if (char) {
          // Line breaking rules in CJK (Kinsoku Shori)
          if (regexCannotEnd.exec(nextChar) || regexCannotStart.exec(char)) {
              return true;
          }
      }
      return false;
  };
  const trimByKinsokuShorui = (prev) => {
      const next = [...prev];
      const prevLine = next[next.length - 2];
      if (!prevLine)
          return prev;
      const lastChar = prevLine[prevLine.length - 1];
      next[next.length - 2] = prevLine.slice(0, -1);
      next[next.length - 1] = lastChar + next[next.length - 1];
      return next;
  };

  const LATIN_REGEX = /[a-zA-Z0-9\u00C0-\u00D6\u00D8-\u00f6\u00f8-\u00ff!"#$%&'()*+,-./:;]/;
  const canBreakInLastChar = (char) => {
      if (char && LATIN_REGEX.test(char))
          return false;
      return true;
  };
  const trimToBreakable = (prev) => {
      // @ts-ignore
      const isBreakingSpace = PIXI.TextMetrics.isBreakingSpace;
      const next = [...prev];
      const prevLine = next[next.length - 2];
      const index = findBreakableIndex(prevLine);
      if (index === -1 || !prevLine)
          return next;
      const trimmedChar = prevLine.slice(index, index + 1);
      const isTrimmedWithSpace = isBreakingSpace(trimmedChar);
      const trimFrom = index + 1;
      const trimTo = index + (isTrimmedWithSpace ? 0 : 1);
      next[next.length - 1] += prevLine.slice(trimFrom, prevLine.length);
      next[next.length - 2] = prevLine.slice(0, trimTo);
      return next;
  };
  const sumTextWidthByCache = (text, cache) => {
      return text.split('').reduce((sum, c) => {
          if (!cache[c])
              throw Error('cannot count the word without cache');
          return sum + cache[c];
      }, 0);
  };
  const findBreakableIndex = (line) => {
      for (let i = line.length - 1; i >= 0; i--) {
          if (!LATIN_REGEX.test(line[i]))
              return i;
      }
      return -1;
  };

  // @ts-ignore
  PIXI.TextMetrics.wordWrap = function (text, { wordWrapWidth, letterSpacing }, canvasTarget) {
      const canvas = canvasTarget || this._canvas;
      const context = canvas.getContext('2d');
      const maxWidth = wordWrapWidth + letterSpacing;
      let lines = [];
      let currentIndex = 0;
      let currentWidth = 0;
      const cache = {};
      const calcWidth = (char) => {
          return this.getFromCache(char, letterSpacing, cache, context);
      };
      runes_1(text).forEach((char, i) => {
          const prevChar = text[i - 1];
          const nextChar = text[i + 1];
          const width = calcWidth(char);
          if (this.isNewline(char)) {
              currentIndex++;
              currentWidth = 0;
              lines[currentIndex] = '';
              return;
          }
          if (currentWidth > 0 && currentWidth + width > maxWidth) {
              currentIndex++;
              currentWidth = 0;
              lines[currentIndex] = '';
              if (this.isBreakingSpace(char)) {
                  return;
              }
              if (!canBreakInLastChar(char)) {
                  lines = trimToBreakable(lines);
                  currentWidth = sumTextWidthByCache(lines[currentIndex] || '', cache);
              }
              if (shouldBreakByKinsokuShorui(char, nextChar)) {
                  lines = trimByKinsokuShorui(lines);
                  currentWidth += calcWidth(prevChar || '');
              }
          }
          currentWidth += width;
          lines[currentIndex] = (lines[currentIndex] || '') + char;
      });
      return lines.join('\n');
  };

})));
//# sourceMappingURL=pixi-cjk.js.map
